{
    "rebuild" : {
        "accessIPv4" : "%(access_ip_v4)s",
        "accessIPv6" : "%(access_ip_v6)s",
        "imageRef" : "%(uuid)s",
        "name" : "%(name)s",
        "key_name" : "%(key_name)s",
        "description" : "%(description)s",
        "adminPass" : "%(pass)s",
        "metadata" : {
            "meta_var" : "meta_val"
        }
    }
}
