{
    "service": {
        "id": "e81d66a4-ddd3-4aba-8a84-171d1cb4d339",
        "binary": "nova-compute",
        "disabled_reason": "maintenance",
        "host": "host1",
        "state": "up",
        "status": "disabled",
        "updated_at": "%(strtime)s",
        "forced_down": false,
        "zone": "nova"
    }
}