{
    "keypair": {
        "fingerprint": "%(fingerprint)s",
        "name": "%(keypair_name)s",
        "private_key": "%(private_key)s",
        "public_key": "%(public_key)s",
        "user_id": "fake"
    }
}
