{
    "networks": [
        {
            "bridge": null,
            "bridge_interface": null,
            "broadcast": null,
            "cidr": null,
            "cidr_v6": null,
            "created_at": null,
            "deleted": null,
            "deleted_at": null,
            "dhcp_start": null,
            "dns1": null,
            "dns2": null,
            "gateway": null,
            "gateway_v6": null,
            "host": null,
            "id": "%(id)s",
            "injected": null,
            "label": "private",
            "multi_host": null,
            "netmask": null,
            "netmask_v6": null,
            "priority": null,
            "project_id": null,
            "rxtx_base": null,
            "updated_at": null,
            "vlan": null,
            "vpn_private_address": null,
            "vpn_public_address": null,
            "vpn_public_port": null,
            "mtu": null,
            "dhcp_server": null,
            "enable_dhcp": null,
            "share_address": null
        }
    ]
}
