{
    "tenant_usages": [
        {
            "start": "%(strtime)s",
            "stop": "%(strtime)s",
            "tenant_id": "6f70656e737461636b20342065766572",
            "total_hours": 1.0,
            "total_local_gb_usage": 1.0,
            "total_memory_mb_usage": 512.0,
            "total_vcpus_usage": 1.0
        }
    ]
}
