{
    "limits": {
        "absolute": {
            "maxServerMeta": 128,
            "maxTotalCores": 20,
            "maxTotalInstances": 10,
            "maxTotalKeypairs": 100,
            "maxTotalRAMSize": 51200,
            "maxServerGroups": 10,
            "maxServerGroupMembers": 10,
            "totalCoresUsed": 0,
            "totalInstancesUsed": 0,
            "totalRAMUsed": 0,
            "totalServerGroupsUsed": 0
        },
        "rate": []
    }
}
