# Copyright (c) 2014 Rackspace Hosting
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import Column
from sqlalchemy import MetaData
from sqlalchemy import Table
from sqlalchemy import Text


def upgrade(engine):
    meta = MetaData()
    meta.bind = engine

    # Drop the compute_node_stats table and add a 'stats' column to
    # compute_nodes directly.  The data itself is transient and doesn't
    # need to be copied over.
    table_names = ('compute_node_stats', 'shadow_compute_node_stats')
    for table_name in table_names:
        table = Table(table_name, meta, autoload=True)
        table.drop()

    # Add a new stats column to compute nodes
    table_names = ('compute_nodes', 'shadow_compute_nodes')
    for table_name in table_names:
        table = Table(table_name, meta, autoload=True)
        stats = Column('stats', Text, default='{}')
        table.create_column(stats)
