{
    "snapshot": {
        "volume_id": "%(volume_id)s",
        "create_info": {
            "snapshot_id": "%(snapshot_id)s",
            "type": "%(type)s",
            "new_file": "%(new_file)s",
            "id": "421752a6-acf6-4b2d-bc7a-119f9148cd8c"
        }
    }
}
