#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import MetaData, Table, func, null, select
from sqlalchemy.sql import and_

from nova import exception
from nova.i18n import _


def upgrade(migrate_engine):
    meta = MetaData(migrate_engine)
    services = Table('services', meta, autoload=True)
    # Count non-deleted services where uuid is null.
    count = select([func.count()]).select_from(services).where(and_(
        services.c.deleted == 0,
        services.c.uuid == null())).execute().scalar()
    if count > 0:
        msg = _('There are still %(count)i unmigrated records in '
                'the services table. Migration cannot continue '
                'until all records have been migrated. Run the '
                '"nova-manage db online_data_migrations" routine.') % {
            'count': count}
        raise exception.ValidationError(detail=msg)
