{
    "instanceAction": {
            "action": "stop",
            "instance_uuid": "%(uuid)s",
            "request_id": "%(request_id)s",
            "user_id": "%(user_id)s",
            "project_id": "%(project_id)s",
            "start_time": "%(strtime)s",
            "message": null,
            "events": [
                {
                    "event": "compute_stop_instance",
                    "start_time": "%(strtime)s",
                    "finish_time": "%(strtime)s",
                    "result": "Success",
                    "traceback": null
                }
            ]
    }
}
