{
    "server": {
        "accessIPv4": "%(access_ip_v4)s",
        "accessIPv6": "%(access_ip_v6)s",
        "OS-DCF:diskConfig": "AUTO",
        "name": "new-server-test",
        "description": "Sample description"
    }
}
